# Suppress nasty future warnings about pandas datetime conersion
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from PASDy.Files import *
from PASDy.Spatiotemp import *
from PASDy.Misc import *

#
from datetime import datetime, timedelta
import pytz
import dateutil
from dateutil.relativedelta import *
import pandas as pd
# Future Warning: mpl needs to be registered
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()
#
pd.options.mode.chained_assignment = None
import sys
import requests
import re
import configparser
import os
import numpy as np
import matplotlib
from matplotlib import pyplot as plt
import matplotlib.backends.backend_pdf
import simplekml
from glob import glob

#JJ packages used for calculating nearest neighbors
#from heapq import nsmallest, nlargest
#from scipy import stats
#from math import sin, cos, sqrt, atan2, radians
#import itertools
#from sklearn.neighbors import KDTree

# Overwrite print() with flush, i.e., no buffer
print = print_flush
#_orig_print = print
#def print(*args, **kwargs):
#    _orig_print(*args, flush=True, **kwargs)
    
#JJ recalculate strange coordinates to proper decimals 
#def deg100min2dec(coord):
#    coord = coord/100
#    deg = np.floor(coord)
#    minute = (coord - deg)/0.6
#    dec = deg+minute
#    return(dec)
    
    
#JJ use x nearest neighbors for averaging
## find k closest lat/lon to every lat/lon coordinates
#def distance(pointA, pointB):
#    return sqrt((pointB[0] - pointA[0]) ** 2 + (pointB[1] - pointA[1]) ** 2)

import zipfile
import os


   
      


def download(archive_file='', source='FTP', server=None, user=None, pswd=None,
             ftp_prefix='', ftp_suffix='', sd_path='', sd_prefix='', sd_suffix=''):

        # if a folder was given
        if not archive_file.endswith('zip'):
            archive_file = os.path.join(archive_file,'') + 'archive.zip'
            
        # Check what has already been downloaded
        archive = zipfile.ZipFile(archive_file, 'a', zipfile.ZIP_DEFLATED)
        archive_list = archive.namelist()
        archive_size = os.stat(archive_file).st_size /1024/1024
        print("%5s files were already downloaded (%.2f MB)." % (len(archive_list), archive_size))

        if source=='FTP':
            # Select new files from FTP
            import ftplib
            remote = ftplib.FTP(server) 
            remote.login(user, pswd)
            remote_list = list(filter(lambda item: item.startswith(ftp_prefix), remote.nlst()))
            print("%5s files found on the FTP server." % len(remote_list))
        
        elif source=='SD':
            # Select new files from SD-Card
            #remote_list = list(filter(lambda item: item.endswith(suffix=''), os.listdir(sd_path)))
            remote_list = [os.path.basename(x) for x in glob(sd_path+'/'+sd_prefix+'*'+sd_suffix)]
            print("%5s files found on the SD backup." % len(remote_list))
        
        else:
            print('Method unknown.')
            return
        
        update_list = list(filter(lambda item: not item in archive_list, remote_list))
        print("%5s new files selected for download." % len(update_list))

        if len(update_list) == 0: return
        
        # Download files and add to Zip
        from io import BytesIO
        i = 0
        for filename in update_list:
            i += 1
            print("  Archiving files %5s/%s: %s" % (i, len(update_list), filename))

            if source=='FTP':
                memory = BytesIO()
                remote.retrbinary("RETR " + filename, memory.write) # FTP->memory
                archive.writestr(filename, memory.getvalue())    # memory->archive
                
            elif source=='SD':
                archive.write(os.path.join(sd_path,'') + filename, filename)

        archive.close()
        if source=='FTP': remote.close()
        file_save_msg(archive_file, "\nArchive", end="\n")
        



# Neutron monitor class
class NM:

    # X = NM()
    def __init__(self, station=None, resolution=None, file=None, config=None, verbose=False):
        
        self.verbose = verbose
        self.data = None
        self.most_recent = False
        
        if config is None:
            self.__config_file = os.path.join(os.path.dirname(__file__), 'config/config.cfg')
            if self.verbose: print('Reading NM config: ' + self.__config_file)
            self.config = configparser.ConfigParser()   
            temp = self.config.read(self.__config_file)
        else:
            self.config = config
        
        # Setup data directory
        if not os.path.exists(self.config['correction']['NM_path']):
            os.makedirs(self.config['correction']['NM_path'])
        
        if station is None:
            self.station = self.config['correction']['NM_station']
        else:
            self.station = station
            
        if resolution is None:
            self.resolution = self.config['correction']['NM_resolution']
        else:
            self.resolution = resolution
        
        if self.resolution == '1day':
            self.resolution_min = 1440
        elif self.resolution == '1hour':
            self.resolution_min = 60
        elif self.resolution == '10min':
            self.resolution_min = 10
        elif self.resolution == '1min':
            self.resolution_min = 1
        
        self.file = file
        #self.file = _filename(datetime.now().strftime('%Y%m%d%H%M%S'))

            
    def _filename(self, string=''):
        if string: string = '-' + str(string)
        return("%s%s-%s%s.dat" \
            % (self.config['correction']['NM_path'], self.station, self.resolution, string))
        
            
    def get(self, start=None, end=None):
        if end is None:
            end = datetime.now()
            self.most_recent = True
        
        # If the user set a filename, just read from it
        if not self.file is None:
            self._reader(file)
            self.data = self.data.tz_localize('UTC')
            return(self)
        
        if start is None:
            print('I do not know which date to look up, please set a start date!')
            return(self)
        
        # Read data
        data = []
        if self.resolution == '1day':
            data.append(self._read(datetime(year=2006, month=1, day=1, tzinfo=pytz.UTC), datetime.datetime.utcnow().date(), self._filename()))
            
        elif self.resolution == '1hour':
            for y in range(end.year - start.year +1):
                data.append(self._read(datetime(year=start.year+y, month=1, day=1, tzinfo=pytz.UTC),
                                       datetime(year=start.year+y, month=12, day=31, hour=23, minute=59, tzinfo=pytz.UTC),
                                       self._filename(start.year+y)))
        
        elif self.resolution == '10min':
            delta = end - start
            for d in [start + relativedelta(months=+m) for m in range(0, delta.days//12+1)]:
                data.append(self._read(d.replace(day=1, hour=0, minute=0),
                                       d.replace(day=1, hour=23, minute=59, second=0) + relativedelta(months=+1) - timedelta(days=1),
                                       self._filename(d.strftime('%Y%m'))))
        
        elif self.resolution == '1min':
            delta = end - start
            for d in [start + timedelta(days=d) for d in range(0, delta.days+2)]:
                data.append(self._read(d.replace(hour=0, minute=0), d.replace(hour=23, minute=59, second=0), self._filename(d.strftime('%Y%m%d'))))
        
        self.data = pd.concat(data)
        #print("Available NM data from: %s to %s" % (self.data.index.min().strftime('%Y-%m-%d %H:%M'), self.data.index.max().strftime('%Y-%m-%d %H:%M')))
        return(self)
    
    # Read existing file or download
    def _read(self, start, end, file):
        
        is_tz_aware = start.tzinfo is not None and start.tzinfo.utcoffset(start) is not None
        if os.path.exists(file):
            
            data = self._reader(file, localize=is_tz_aware)
            
            now = datetime.now(pytz.utc)
            if end > now: self.most_recent = True 
            if start > now: return(None) # None when looking for a future day
            
            #print(data.index)
            if data.index.min() <= start and (data.index.max() + timedelta(minutes=self.resolution_min) >= end or self.most_recent):
                if not is_tz_aware:
                    data = data.tz_localize('UTC')
                return(data)
            else:
                print("\nData period is not covering the requested period.")
                print("     Data from: %s to %s" % (data.index.min().strftime('%Y-%m-%d %H:%M'), data.index.max().strftime('%Y-%m-%d %H:%M')))
                print("Requested from: %s to %s" % (start.strftime('%Y-%m-%d %H:%M'), end.strftime('%Y-%m-%d %H:%M')))
        
        self._download(start, end, file)
        return(self._reader(file))
            
                
    def _reader(self, file, localize=True):
        print('.', end='')
        data = pd.read_csv(file, index_col=0, header=None, sep=";", names=['time','NM'],
                           parse_dates=True, infer_datetime_format=True)
        if localize: 
            data = data.tz_localize('UTC')
        return(data)
        
        
    
    def _download(self, start, end, file, force=True):
    
        force_str = '&force=1' if force else ''
        url = "http://www.nmdb.eu/nest/draw_graph.php?wget=1&stations[]=%s&tabchoice=revori&dtype=corr_for_efficiency&tresolution=%i%s&date_choice=bydate&start_year=%s&start_month=%s&start_day=%s&start_hour=%s&start_min=%s&end_year=%s&end_month=%s&end_day=%s&end_hour=%s&end_min=%s&yunits=0" \
            % (self.station, self.resolution_min, force_str,
            start.strftime('%Y'), start.strftime('%m'), start.strftime('%d'), start.strftime('%H'), start.strftime('%M'),
            end.strftime('%Y'), end.strftime('%m'), end.strftime('%d'), end.strftime('%H'), end.strftime('%M'))
        if self.verbose: print('Request URL: ' + url)
            
        print("\n  Downloading... ", end='')
        r = requests.get(url)
        if self.verbose: print('received %s chars... ' % len(r.text), end='')
        
        # Write into file
        re_data_line = re.compile(r'^\d')
        with open(file, 'w') as f:
            for line in r.text.splitlines():
                if re_data_line.search(line):
                    f.write(line + "\n")
        file_save_msg(file, '')

        
#def file_save_msg(file, name='File', end=''):
#    size = os.stat(file).st_size
#    if   size < 1024:    sizestr = "%.0f Bytes" % (size)
#    elif size < 1024**2: sizestr = "%.0f KB"    % (size/1024)
#    else:                sizestr = "%.1f MB"    % (size/1024/1024)
#    print("%s saved to %s (%s)" % (name, file, sizestr), end=end)
    

#def ifdef(a, b, convert=float):
#    if not a:
#        return(b)
#    else:
#        return(convert(a))

#def xsplit(s, range=0, type=str):
#    a = np.array(s.replace(',',' ').split(), dtype=type)
    # fill up with NaNs
#    for i in np.arange(range-len(a)):
#        a = np.append(a, np.nan)
#    return(a)

re_crnsdata = re.compile(r'^\d+\,')
re_dataselect = re.compile(r'//DataSelect\s*=\s*(\w+)')
re_columns = re.compile(r'/+(RecordNum.+)')
re_columns2 = re.compile(r'/+(GpsUTC.+)')

def read(filename, archive=None, tz='UTC'):
    #print(archive)
#    if not archive is None:
#        archive = zipfile.ZipFile(archive_file, 'r')

    header_skip = 0
#    data_columns = ''
    datastr = ''
#    debug = 0
    with archive.open(filename,'r') if archive else open(filename, encoding="cp850") as file:
        for line in file:
#            debug += 1
#            if debug >= 1:
#                print(line)
            if isinstance(line, bytes):
                line = line.decode(errors='ignore')
            if re_crnsdata.search(line):
                datastr += line
#            else:
#                header_skip += 1
#                match_dselect = re_dataselect.search(line)
#                if match_dselect:
#                    data_select = match_dselect.group(1)
#                    continue
#                match_columns = re_columns.search(line)
#                if match_columns:
#                    data_columns = match_columns.group(1)
#                    continue
#                match_columns2 = re_columns2.search(line)
#                if match_columns2:
#                    data_columns += ',' + match_columns2.group(1)
#                    continue
                    
    #print('Parsed', header_skip, 'lines of auxiliary information.')
#    if not data_columns:
#        print('(No header in file %s)' % filename, end='')
#        return(None)
    
    #data_columns = data_columns.replace(',,',',')
#    data_columns = re.sub(r'\s*,+\s*', ',', data_columns)
#    data_columns = data_columns.split(',')
                
    # Read that file
#    filename_orig = filename
#    if archive:
#        filename = archive.open(filename,'r')
#    try:
#        data = pd.read_csv(filename, sep=',', skiprows=header_skip, names=data_columns,
#                    index_col=1, parse_dates=True, infer_datetime_format=True, encoding='cp850',
#                    error_bad_lines=False, warn_bad_lines=False)
#    except:
#        print("(Bad lines in file %s)" % filename_orig, end='')
#        return(None)
                    
    # Tidy up
#    data = data.loc[data.index.dropna()]
    
    ## Find out which column names are null
#    bad_columns = []
#    for c in range(len(data_columns)):
#        if pd.isnull(data_columns[c]):
#            bad_columns.append(c)
    ## Reassign data using only non-bad columns
#    data = data.iloc[:, [i for i,n in enumerate(data.columns) if i not in bad_columns]]
    
    return(datastr+"\n")


#def chomp(x):
#    if x.endswith("\r\n"): return x[:-2]
#    if x.endswith("\n") or x.endswith("\r"): return x[:-1]
#    return(x)
    
def read_header(filename, archive=None):

    import re
    import os
    re_crnsdata = re.compile(r'^\d+\,')
    re_columns = re.compile(r'/+(RecordNum.+)')
    re_columns2 = re.compile(r'/+(GpsUTC.+)')
    data_columns = []

    with archive.open(filename,'r') if archive else open(filename) as file:
        for line in file:
            if isinstance(line, bytes):
                line = line.decode(errors='ignore')
                
            if re_crnsdata.search(line):
                break
            else:
                match_columns = re_columns.search(line)
                if match_columns:
                    data_columns = match_columns.group(1)
                    continue
                match_columns2 = re_columns2.search(line)
                if match_columns2:
                    data_columns += ',' + match_columns2.group(1)
                    continue

    if len(data_columns)<=0:
        print('Input file contains no headers (e.g., //RecordNum...)!')
        return(None)
    else:
        data_columns = chomp(data_columns)
        data_columns = re.sub(r'\s*,+\s*', ',', data_columns)
        data_columns = re.sub(r',+', ',', data_columns)
        data_columns = data_columns.split(',')
        return(data_columns)
        
import io

def make_pandas(datastr, columns=None, header_skip=0):
    
    try:
        data = pd.read_csv(io.StringIO(datastr), sep=',', skiprows=header_skip, names=columns,
                            index_col=1, parse_dates=True, infer_datetime_format=True, encoding='cp850',
                            error_bad_lines=False, warn_bad_lines=False, skipinitialspace=True)
    except Exception as e:
        print("ERROR interpreting data format: "+str(e))
        return(None)
                    
    # Tidy up
    data = data.loc[data.index.dropna()]
    len1 = len(data)
    data = data.sort_index().drop_duplicates().tz_localize('UTC')
    len2 = len(data)
    if len2 < len1: print("\n  Removed %i duplicates." % (len1-len2), end='')
    
    ## Find out which column names are null
    bad_columns = []
    for c in range(len(columns)):
        if pd.isnull(columns[c]):
            bad_columns.append(c)
    ## Reassign data using only non-bad columns
    data = data.iloc[:, [i for i,n in enumerate(data.columns) if i not in bad_columns]]
    return(data)
    
    
# Save data as KML
def save_KML(data, var, lat, lon, file='test', format_str="%.0f%%", collim=None, cmap='Spectral_r'):
    
    data.loc[data[lat]<=0, lat] = np.nan
    data.loc[data[lon]<=0, lon] = np.nan
    data = data.dropna(subset=[var, lon, lat])

    if collim is None:
        xmin = np.min(data.loc[:,var])
        xmax = np.max(data.loc[:,var])
    else:
        xmin = float(collim[0])
        xmax = float(collim[1])
    
    data.loc[:,'color'] = None
    cmap = matplotlib.cm.get_cmap(cmap)
    for i, row in data.iterrows():
        tmp = (data.at[i,var]-xmin)/(xmax-xmin)
        data.at[i,'color'] = matplotlib.colors.rgb2hex(cmap(tmp)[:3])
    
    kml = simplekml.Kml()
    
    for i, row in data.iterrows():
        label = format_str % row[var] if format_str else ''
        pnt = kml.newpoint(name=label, coords=[(row[lon],row[lat])])
        pnt.style.iconstyle.color = row['color'].replace('#','#ff')
        pnt.style.iconstyle.scale = 1
        pnt.style.iconstyle.icon.href = 'http://maps.google.com/mapfiles/kml/pal2/icon18.png'
        pnt.style.labelstyle.scale = 0.8

    kml.save(file + '.kml')


# Figures

class Fig:
    fig = None # fig object given by pdfpages
    ax = None  # current single axis
    time_format = '%Y'
    layout = (1,1)
    axi = 1
    submode = False

    def __init__(self, fig=None, title='', layout=(1,1), ylabel='', size=(11.69,8.27), ylim=None):

        # Single PDF page when fig is provided
        if not fig is None:
            Fig.fig = fig
            plt.figure(figsize=size)
            Fig.layout = layout
            Fig.axi = 0
        
        if layout != (1,1):
            # For complex layout, do not do anything, just wait for next "with Fig()"
            Fig.submode = True
        else:
            Fig.axi += 1
            Fig.ax = plt.subplot("%i%i%i" % (Fig.layout[0], Fig.layout[1], Fig.axi))
            
            plt.title(title)
            Fig.ax.set(xlabel='', ylabel=ylabel)
            if ylim: Fig.ax.set_ylim(ylim)
            for a in ("top", "right"):
                Fig.ax.spines[a].set_visible(False)
                Fig.ax.get_xaxis().tick_bottom()
                Fig.ax.get_yaxis().tick_left()
            Fig.ax.grid(b=True, alpha=0.2)
            time_format = re.sub(r'(\w)', r'%\1', Fig.time_format)
            Fig.ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter(time_format.replace('\\','')))
            plt.tick_params(labelsize=9)
    
    # Entering `with` statement
    def __enter__(self):
        return(Fig.ax) # makes possibe: with Fig() as ax: ax.change
    
    # Exiting `with` statement    
    def __exit__(self, type, value, traceback):
        # deactivate submode if axis counter exceeds layout shape
        if Fig.submode:
            if Fig.axi == Fig.layout[0]*Fig.layout[1]:
                Fig.submode = False
        else:
            # save and close PDF page
            Fig.fig.savefig(bbox_inches='tight')
            plt.close()
            
#def yesno2bool(s):
#    s = s.lower()
#    if s == 'yes' or s == 'y':
#        return(True)
#    else:
#        return(False)

config = None

def main(configfile=None):
    print("Welcome to instant-PASDy!")

    # Config
    config = configparser.ConfigParser()
    if configfile is None: configfile = 'config.cfg'
    if not os.path.isfile(configfile):
        print('Config file required! Either specify as argument or provide config.cfg')
        return()
    else:
        print("Working with %s" % configfile)
            
    temp = config.read(configfile)
    
    # clean config
    if os.path.isdir(config['input']['path']):
        # add trailing slash
        config['input']['path'] = os.path.join(config['input']['path'], '')
    # check whether output path exists
    if not os.path.exists(config['output']['out_path']):
        os.makedirs(config['output']['out_path'])
    # add trailing slash
    config['output']['out_path'] = os.path.join(config['output']['out_path'], '') 
    
    if 'update' in config and config['update']['data_source'] in ['FTP','SD']:
        print('Updating data from ' + config['update']['data_source'] + ' into ' + config['input']['path'])
        download(archive_file=config['input']['path'], source='SD', #config['update']['data_source'],
            server=config['update']['FTP_server'], user=config['update']['FTP_user'],
            pswd=config['update']['FTP_pswd'], ftp_prefix=config['update']['FTP_prefix'],
            ftp_suffix=config['update']['FTP_suffix'], sd_path=config['update']['SD_path'],
            sd_prefix=config['update']['SD_prefix'], sd_suffix=config['update']['SD_suffix'])
        download(archive_file=config['input']['path'], source='FTP', #config['update']['data_source'],
            server=config['update']['FTP_server'], user=config['update']['FTP_user'],
            pswd=config['update']['FTP_pswd'], ftp_prefix=config['update']['FTP_prefix'],
            ftp_suffix=config['update']['FTP_suffix'], sd_path=config['update']['SD_path'],
            sd_prefix=config['update']['SD_prefix'], sd_suffix=config['update']['SD_suffix'])

    
    # Input
    print('Reading data from ' + config['input']['path']+' ', end='')
    
    pattern = config['input']['path']+config['input']['prefix']+'*'+config['input']['suffix']
    archive = None
    files = []
    if config['input']['path'].endswith('.zip'):
        archive = zipfile.ZipFile(config['input']['path'], 'r')
        files = [x for x in archive.namelist() if x.endswith(config['input']['suffix']) and x.startswith(config['input']['prefix'])]
    elif os.path.isdir(config['input']['path']):
        files = glob(pattern)
    else:
        if 'gui' in globals():
            gui.Popup("Input path is neither a folder nor a zip archive.", title='Read ERROR', non_blocking=True)
        else:
            print("ERROR: Input path is neither a folder nor a zip archive.")
        return(None)
    
    datastr = ''
    data_columns = []
    i = 0
    for filename in files:
#        print('.', end='')
        i += 1
        pc = i/len(files)*100
        if pc%10==0: print("%.0f%%.." % pc, end='')
        if datastr=='':
            data_columns = read_header(filename, archive)
            datastr = read(filename, archive)
        else:
            datastr += read(filename, archive) #data.append() #, verify_integrity=True)
    
    if not archive is None: archive.close()
    #DEBUG output raw data file
    #with open('temp.txt','a') as f:
    #    f.write(datastr)
    data = make_pandas(datastr, data_columns)
    
    N = config['correction']['new_neutron_column']
    tres = config['input']['resolution_column']
    lat = config['input']['latitude_column']
    lon = config['input']['longitude_column']
    p  = config['input']['pressure_column']
    rh = config['input']['humidity_column']
    T  = config['input']['temperature_column']
    ah = config['correction']['new_humidity_column']
    inc = config['correction']['new_incoming_column']
        
    # check wrong column names
    for name in [p, rh, T]:
        if not name in data.columns:
            print('Warning: %s is not a column name.' % name)
            data[name] = np.nan
    
    if not tres or not tres in data.columns:
        tres = 'N1ET_sec'
        if 'resolution_default' in config['input'] and config['input']['resolution_default']:
            data[tres] = int(config['input']['resolution_default'])
        else:
            data[tres] = 60 # set 60 sec as the fallback time resolution
    
    ## Sum up neutron columns and convert to units of cph
    data[N] = 0
    for c in xsplit(config['input']['neutron_columns']):
        data[N] += data[c] / data[tres] * 3600
    
    #data[['N1Cts','N2Cts',N]].to_csv('temp.csv')
    
    # Thermal neutrons
    if config['input']['thermal_neutron_columns']:
        Nth = N+'_thermal'
        data[Nth] = 0
        for c in xsplit(config['input']['thermal_neutron_columns']):
            data[Nth] += data[c] / data[tres] * 3600
    else:
        Nth = None
    
    print('\n  Data period from', str(data.index.min()), 'to', str(data.index.max()))

    # Clean
    
    print('Cleaning... ', end='')
    (Na,Nb)  = xsplit(config['clean']['neutron_range'], range=2, type=float)
    (pa,pb)  = xsplit(config['clean']['pressure_range'], range=2, type=float)
    (rha,rhb)= xsplit(config['clean']['humidity_range'], range=2, type=float)
    (Ta,Tb)  = xsplit(config['clean']['temperature_range'], range=2, type=float)
    
    if config['clean']['invalid_data'] == 'drop lines':
        len_before = len(data)
        data = data.query("@N > %s and @N < %s" % (Na, Nb))
        data = data.query("@p > %s and @p < %s" % (pa, pb))
        data = data.query("@rh> %s and @rh< %s" % (rha,rhb))
        data = data.query("@T > %s and @T < %s" % (Ta, Tb))
        if lat in data.columns: data = data.query("%s != 0.0" % lat)
        if lon in data.columns: data = data.query("%s != 0.0" % lon)
        print("dropped %s invalid lines." % (len_before - len(data)))
    else:
        nan_before = data.isnull().sum().sum()
        data.loc[(data[N] < Na ) | (data[N] > Nb) , N ] = np.nan
        data.loc[(data[p] < pa ) | (data[p] > pb) , p ] = ifdef(config['clean']['missing_pressure'], np.nan)
        data.loc[(data[rh]< rha) | (data[rh]> rhb), rh] = ifdef(config['clean']['missing_humidity'], np.nan)
        data.loc[(data[T] < Ta ) | (data[T] > Tb) , T ] = ifdef(config['clean']['missing_temperature'], np.nan)
        if lat in data.columns: data.loc[data[lat] == 0.0, lat] = np.nan
        if lon in data.columns: data.loc[data[lon] == 0.0, lon] = np.nan
        print("replaced %s values by NaN." % (data.isnull().sum().sum() - nan_before))
    
    #JJ only for Jülich rover necessary - strangely formatted coordinates    
    if config['input']['suffix'] == '.FZJ':
        data[lat]= pd.to_numeric(data[lat].astype(str).replace({'N':''}, regex=True), errors='coerce') # replace N if existent in column and convert to numeric
        data[lon]= pd.to_numeric(data[lon].astype(str).replace({'E':''}, regex=True), errors='coerce') # replace E if existent in column and convert to numeric
        data[lon] = deg100min2dec(data[lon]) # convert to readable coordinate
        data[lat] = deg100min2dec(data[lat])
    
    if yesno2bool(config['clean']['interpolate_neutrons']):
        data[N] = data[N].interpolate()
        if Nth: data[Nth] = data[Nth].interpolate()
        
    # Incoming
    print('Reading incoming radiation data from ' + config['correction']['NM_path'], end='')
    M = NM(config=config).get(start=data.index.min(), end=data.index.max())
    M.data = M.data.reindex(data.index, method='nearest', limit=1, tolerance=timedelta(seconds=M.resolution_min*60/2-1)).interpolate()

    # Corrections
    print("\nCorrections... ", end='')
    data[inc] = M.data
    data[ah] = 6.112*np.exp(17.67*data[T]/(243.5+data[T]))/(273.15 + data[T]) * 2.1674 * data[rh]
    data['correct_h'] = 1 + float(config['correction']['alpha']) * (data[ah] - float(config['correction']['humidity_ref']))
    data['correct_p'] = np.exp((data[p] - float(config['correction']['pressure_ref'])) / float(config['correction']['beta']))
    data['correct_inc'] = 1/(1 - float(config['correction']['gamma']) * (1 - data[inc] / float(config['correction']['incoming_ref']))) 
    data[N] = data[N] * data['correct_p'] * data['correct_h'] * data['correct_inc']
    if Nth: data[Nth] = data[Nth] * data['correct_p'] * data['correct_h'] * data['correct_inc']
    print("applied for pressure (%+.0f%%), abs. humidity (%+.0f%%), and incoming radiation (%+.0f%%)." % \
    (100*(data['correct_p'].mean()-1), 100*(data['correct_h'].mean()-1), 100*(data['correct_inc'].mean()-1)))

    # Aggregate
    if config['clean']['aggregate']:
        print('Aggregate data to '+config['clean']['aggregate']+' time steps.')
        data_orig = data.copy() 
        window = config['clean']['aggregate'].replace('s','')
        window = window.replace('hour','H').replace('min','Min').replace('day','D').replace('week','W').replace('month','M').replace('year','A')
        aggobj = data.resample(window)
        data = aggobj.mean()
        data[N+'_std'] = aggobj.std()[N]
        if Nth: data[Nth+'_std'] = aggobj.std()[Nth]
    
    ## Smooth
    if config['clean']['smooth']:
        print('Smooth data by a moving average over '+config['clean']['smooth']+' time steps.')
        rollobj = data[N].rolling(int(config['clean']['smooth']), center=True)
        data[N+'_smooth'] = rollobj.mean()
        data[N+'_smooth'+'_std'] = rollobj.std()
        N = N+'_smooth'
        if Nth:
            rollobj = data[Nth].rolling(int(config['clean']['smooth']), center=True)
            data[Nth+'_smooth'] = rollobj.mean()
            data[Nth+'_smooth'+'_std'] = rollobj.std()
            Nth = Nth+'_smooth'
            
    ##JJ smooth with nearest neighbors
    if config['clean']['number_locations']:
        print('Smooth data with '+config['clean']['number_locations']+' nearest neighbors.')
        data = data.dropna(subset=['LatDec', 'LongDec']) # drop rows without Latitude or Longitude information
        maxdist, RecNums = Neighbours(data[lat], data[lon], int(config['clean']['number_locations']), data)
        
        data['maxdistance'] = maxdist
        data['neighbors_RecordNum'] = RecNums.tolist()
        
        test = []
        Recind = []
        res = []
        result = []
        resTh = []
        result_Th = []
        
        for i in range(len(data)): # store the nearest recordnumber in neighbors
            test = data['RecordNum'].iloc[data['neighbors_RecordNum'][i]].values
            Recind.append(test.tolist())
        
        data['neighbors_RecordNum'] = Recind
        
        for i in range(len(data)): # calculate average N from neighbors index
            res = data[N][data['RecordNum'].isin(data['neighbors_RecordNum'][i])].mean()
            result.append(res)
            resTh = data[Nth][data['RecordNum'].isin(data['neighbors_RecordNum'][i])].mean()
            result_Th.append(resTh)
        
        data[N+'_neighbors'] = result
        data[Nth+'_neighbors'] = result_Th
            
    
    if config['conversion']['N2sm_method'] != 'None':     
        
        # Calibrate
        N0 = config['conversion']['N0']
        calibration = False
        if not N0:
            print('Calibration... ', end='')
            calibration = True
            start = dateutil.parser.parse(config['conversion']['campaign_start'])
            end = dateutil.parser.parse(config['conversion']['campaign_end'])
            datac = data.copy()
            datac = datac.loc[start:end]
            Nc = datac.N.mean()
            Nc_error = datac.N.std()
            
            N0 = Nc/(float(config['conversion']['a0']) / (float(config['conversion']['measured_sm'])/float(config['conversion']['bulk_density']) + float(config['conversion']['a2']) + float(config['conversion']['lattice_water']) + float(config['conversion']['soil_org_carbon'])*0.556) + float(config['conversion']['a1']))
            N0_error = N0-(Nc+Nc_error)/(float(config['conversion']['a0']) / (float(config['conversion']['measured_sm'])/float(config['conversion']['bulk_density']) + float(config['conversion']['a2']) + float(config['conversion']['lattice_water']) + float(config['conversion']['soil_org_carbon'])*0.556) + float(config['conversion']['a1']))
            
            print('N0 = %.0f +/- %.0f' % (N0, abs(N0_error)))
        
        # Conversion
        sm = config['conversion']['new_moisture_column']
        sm_neighbors = sm+'_neighbors'
        print('Conversion to water content "%s".' % sm)
        
        while True:
            data[sm] = (float(config['conversion']['a0'])/(data[N]/int(N0)-float(config['conversion']['a1']))-float(config['conversion']['a2']) \
                        -float(config['conversion']['lattice_water']) - float(config['conversion']['soil_org_carbon'])*0.556) * float(config['conversion']['bulk_density'])
            
            if 'N_neighbors' in data.columns:
                #JJ
                data[sm_neighbors] = (float(config['conversion']['a0'])/(data['N_neighbors']/int(N0)-float(config['conversion']['a1']))-float(config['conversion']['a2']) \
                            -float(config['conversion']['lattice_water']) - float(config['conversion']['soil_org_carbon'])*0.556) * float(config['conversion']['bulk_density'])

            if config['conversion']['road_moisture'] and float(config['conversion']['road_width']) > 0 and not 'correct_road' in data.columns:
                print("Corrections for road effect... ", end='')
                #if   config['correction']['road_material'] == 'stone':    road_theta = 0.04
                #elif config['correction']['road_material'] == 'concrete': road_theta = 0.07
                #elif config['correction']['road_material'] == 'asphalt':  road_theta = 0.10
                road_theta = road_moisture
                data['correct_road'] = 1 + 0.42*(1-np.exp(-0.5*float(config['conversion']['road_width']))) * (1.06-4*road_theta-(0.16 + road_theta)/(0.39 + data[sm]))
                data[N] /= data['correct_road']
                print("(%+.0f%%)." % (100*(1/data['correct_road'].mean()-1)))
            else:
                break
        
        (a,b) = xsplit(config['clean']['sm_range'], type=float)
        if config['clean']['invalid_data'] == 'drop':
            data = data.query("@sm > %s and @sm < %s" % (a,b))
        else:
            data.loc[(data[sm] < a) | (data[sm] > b), sm] = np.nan
            if sm_neighbors in data.columns:
                data.loc[(data[sm_neighbors] < a) | (data[sm_neighbors] > b), sm_neighbors] = np.nan

        data[sm+'_%'] = data[sm] * 100
        
    # Output
    
    ## Select a certain period
    if config['clean']['start'] or config['clean']['end']:
        if config['clean']['start']:
            start = dateutil.parser.parse(config['clean']['start'])
        else:
            start = data.index.min()
        if config['clean']['end']:
            end = dateutil.parser.parse(config['clean']['end'])
        else:
            end = data.index.max()
        data = data.loc[start:end]
    
    
    start_str = data.index.min().strftime('%Y%m%d')
    end_str = data.index.max().strftime('%Y%m%d')
    end_str = '-'+end_str if start_str != end_str else '' 
    output_basename = os.path.join(config['output']['out_path'],'') + config['output']['out_basename'] + '-' \
                      + start_str + end_str #config['input']['prefix']
    
    if yesno2bool(config['output']['make_CSV']):
        print('CSV ', end='')
        datetime_format = re.sub(r'(\w)', r'%\1', config['output']['CSV_datetime_format'])
        try:
            data.to_csv(output_basename+'.csv',
                date_format = datetime_format,
                sep = config['output']['CSV_column_sep'].strip('\"'),
                decimal = config['output']['CSV_decimal'],
                float_format = '%' + config['output']['CSV_float_format'],
                na_rep = config['output']['CSV_NaN'])
        except:
            print('ERROR: Cannot write, probably permission denied.')
            
        file_save_msg(output_basename + '.csv', "data", end="\n")
    
    # Figures
    if yesno2bool(config['output']['make_PDF']):

        print('Plotting... ', end='')
        figure_name = output_basename + '.pdf'
        
        #print(data[N].to_string())
        ## With PdfPages: Create PDF on enter and close PDF on exit
        ## see https://matplotlib.org/gallery/misc/multipage_pdf.html
        with matplotlib.backends.backend_pdf.PdfPages(figure_name) as fig:
            
            Fig.time_format = config['output']['PDF_time_format']
            
            # Neutrons
            with Fig(fig, title='Neutron time series, corrected', ylabel='Neutrons (in cph)') as ax:
                if not data[N].isnull().all():
                    plt.plot(data[N], label="N", drawstyle='steps')
                else:
                    plt.plot(data.index[0], 0)

                if 'N_neighbors' in data.columns:
                    #JJ
                    if not data['N_neighbors'].isnull().all():
                        plt.plot(data['N_neighbors'],label="test",drawstyle='steps')

        #            if N+'_std' in data.columns:
        #                ax.fill_between(data.index, data[N]-data[N+'_std'], data[N]+data[N+'_std'], alpha=0.25, lw=0, step='pre')

#            if Nth:
#                with Fig(fig, title='Thermal neutron time series, corrected', ylabel='Thermal neutrons (in cph)') as ax:
#                    if not data[Nth].isnull().all():
#                        plt.plot(data[Nth], label="N_thermal", drawstyle='steps')

            
            # Air properties
            with Fig(fig, layout=(2,2)) as axes:
                if not data[p].isnull().all(): 
                    with Fig(ylabel='Air pressure (in mbar)'):
                        plt.plot(data[p], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=1: axes[1].axis('off')
                
                if not data[T].isnull().all(): 
                    with Fig(ylabel='Air temperature (in $^\circ$C)'):
                        plt.plot(data[T], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=2: axes[2].axis('off')
                
                if not data[rh].isnull().all(): 
                    with Fig(ylabel='Relative humidity (in %)'):
                        plt.plot(data[rh], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=3: axes[3].axis('off')
                
                if not data[ah].isnull().all(): 
                    with Fig(ylabel='Absolute humidity (in g/m$^3$)'):
                        plt.plot(data[ah], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=4: axes[4].axis('off')

            # Moisture
            if config['conversion']['N2sm_method'] != 'None':     
                if not data[sm].isnull().all():
                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content (in %)', ylim=(0,100)):        
                        plt.plot(data[sm+'_%'], drawstyle='steps')
                        if calibration:
                            plt.plot([datac.index.min(), datac.index.max()], [float(config['conversion']['measured_sm'])*100, float(config['conversion']['measured_sm'])*100], color='red', lw=10, alpha=0.5)
                            plt.text(datac.index.min(), float(config['conversion']['measured_sm'])*105, 'calibration', rotation=90, va='bottom', color='red', alpha=0.5)

            # More plots
            for var in xsplit(config['output']['PDF_plots']):
                if var in data.columns:
                    with Fig(fig, title=var, ylabel=var):        
                        if not data[var].isnull().all():
                            plt.plot(data[var], drawstyle='steps')
        
        # Done
        file_save_msg(figure_name, "figures", end="\n")
        
    # Make KML if lat/lon is provided
    if lat in data.columns and lon in data.columns:
    
        if yesno2bool(config['output']['make_KML_neutrons']):
            if not data[N].isnull().all():
                print('Creating_KML... ', end='')
                collim = xsplit(config['output']['KML_neutrons_range'])
                if len(collim)<=1: collim = None
                save_KML(data, N, lat, lon, file=output_basename+'-'+config['correction']['new_neutron_column'], format_str='', cmap='Spectral', collim=collim)#(3000,9000)) # long track: 5000,11000
                file_save_msg(output_basename+'-'+config['correction']['new_neutron_column']+'.kml', '', end="\n")
            
#            if Nth:
#                print("\n"+'Creating_KML... ', end='')
#                save_KML(data, Nth, lat, lon, file=output_basename+'-'+config['correction']['new_neutron_column']+'_thermal', format_str='', cmap='Spectral') #, collim=(3000,9000)) # long track: 5000,11000
#                file_save_msg(output_basename+'-'+config['correction']['new_neutron_column']+'_thermal.kml', '', end='')
        
        if yesno2bool(config['output']['make_KML_sm']):
            if not data[sm].isnull().all():
                print('Creating_KML... ', end='')
                collim = xsplit(config['output']['KML_sm_range'])
                if len(collim)<=1: collim = None
                save_KML(data, sm+'_%', lat, lon, file=output_basename+'-'+sm, format_str='%.0f%%', collim=collim)#(5,80))
                file_save_msg(output_basename+'-'+sm+'.kml', '', end="\n")
        
        if yesno2bool(config['output']['make_KML_other']):
            other = config['output']['KML_other_column']
            if other in data.columns:
                if not data[other].isnull().all():
                    print('Creating_KML... ', end='')
                    collim = xsplit(config['output']['KML_other_range'])
                    if len(collim)<=1: collim = None
                    if yesno2bool(config['output']['KML_other_reverse']):
                        cmap = 'Spectral_r'
                    else:
                        cmap = 'Spectral'
                    save_KML(data, other, lat, lon, file=output_basename+'-'+other, format_str='%'+config['output']['KML_other_format'], cmap=cmap, collim=collim)#(5,80))
                    file_save_msg(output_basename+'-'+other+'.kml', '', end="\n")
            
         # Contourf Interpolation    
#        P = P[P.LatDec>47.829]
#        P = P[P.N_smooth > 0]
#        ngridx = int(10000*(P.LongDec.max()-P.LongDec.min()))
#        ngridy = int(10000*(P.LatDec.max()-P.LatDec.min()))
#        print(ngridx, ngridy)
#        x = [int(10000*x) for x in (P.LongDec-P.LongDec.min())]
#        y = [int(10000*x) for x in (P.LatDec-P.LatDec.min())]
#        z = P.moisture.interpolate().values*100
#        xi = np.linspace(0, ngridx, ngridx)
#        yi = np.linspace(0, ngridy, ngridy)
#        triang = tri.Triangulation(x, y)
#        interpolator = tri.LinearTriInterpolator(triang, z)
#        Xi, Yi = np.meshgrid(xi, yi)
#        zi = interpolator(Xi, Yi) 
#        fig, ax = plt.subplots(figsize=(10,5))
#        plt.contourf(zi, cmap='Spectral', levels=10)
#        plt.scatter(x,y, color='black', s=1)
            
    print("Done.")
    return(data)

# Direct call (instead of import)
if __name__ == "__main__":
    if len(sys.argv) > 1:
        main(sys.argv.pop(1))
        if len(sys.argv) > 1:
            param2 = sys.argv.pop(1)
        else:
            param2 = ''
    else:
        main()
    
    if param2 != '--nonstop':
        print("Press any key to close.", end='')
        input()
    
