# PASDy.Spatiotemp -- Spatiotemporal data analysis

import numpy as np


# Spatial functions

## Distance
##
### use x nearest neighbors for averaging
### find k closest lat/lon to every lat/lon coordinates
### 
def distance(pointA, pointB):
    return sqrt((pointB[0] - pointA[0]) ** 2 + (pointB[1] - pointA[1]) ** 2)


## Coordinate conversion
##
### recalculate strange coordinates to proper decimals 
### 
def deg100min2dec(coord):
    import numpy
    coord = coord/100
    deg = numpy.floor(coord)
    minute = (coord - deg)/0.6
    dec = deg+minute
    return(dec)


# Nearest Neighbor
#
## JJ packages used for calculating nearest neighbors
## 
from heapq import nsmallest, nlargest
from scipy import stats
from math import sin, cos, sqrt, atan2, radians
import itertools
from sklearn.neighbors import KDTree


### From lat, lon get distance in km
###
def distanceinkm(xstand, ystand, xfor, yfor): # approximate radius of earth in km
   
    R = 6373.0
    lat1 = radians(xstand)
    lon1 = radians(ystand)
    lat2 = radians(xfor)
    lon2 = radians(yfor)
    
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    distance = R * c
    
    return(distance)

### 
###
def Neighbours(lat, lon, numpoints, data):

    numpoints = numpoints+1 # points itself included - therefore add one to get desired output
    points = np.vstack((lat, lon)).T # create numpy array from coordinates
    
    kdt = KDTree(points, leaf_size=30, metric='euclidean') # build tree with euclidean distances
    neighborpoints = kdt.query(points, k=numpoints, return_distance=False) # find the closest k points  
        
    # calculate maximal distance within the k clostest coordinates
    neighbordist = []
    for i in range(len(points)):
        furthest = np.array(nlargest(1, points[neighborpoints[i]], lambda p : distance(p, (points[i])))) # find the coordinates for the k closest points
        neighbordist.append(distanceinkm(points[i][0],points[i][1], furthest[0][0],furthest[0][1])*1000)
        
    return neighbordist, neighborpoints

    #neighborpoints = [] # create empty list for all results
    
    #for i in range(len(points)): ## iterate for all points
    #    nearest = np.array(nsmallest(numpoints, points, lambda p : distance(p, (points[i])))) # find the coordinates for the k closest points
    #    print(nearest)
    #    index = [] # create empty list for the index of the positions
               
    #    for m in range(len(nearest)): 
    #        idx, rest = np.where(points==nearest[m]) # find the rows of these closest k coordinates
    #        idx = stats.mode(idx) # calculate the mode - necessary because sometimes several outputs from above
    #        index.append(idx.mode.tolist()) # create list from the k rows found
                
    #    neighborpoints.append(index) # create list of k closest locations for all locations
    #for i in range(len(neighborpoints)):
    #    neighborpoints[i] = list(itertools.chain(*neighborpoints[i])) # reduce for one nested list
    
